import React from 'react';
import styles from './inputs.module.scss';
import { DropdownTrigger, DropdownContainer, DropdownItem, DropdownItemContainer } from '../dropdowns';
import Info from '../info';
import { Tooltip, TooltipContent, TooltipTrigger } from '../tooltip';
import { Popover } from '../layout/Popover';

type SliderProps = {
  label: string;
  type: 'slider';
  value: number;
  min: number;
  max: number;
  step: number;
  showValue?: boolean;
  unit?: string;
  tooltip?: string;
  onChange?: (value: number) => void;
};

type TextProps = {
  label?: string;
  type: 'text';
  size?: 'xs' | 'sm' | 'md' | 'xl';
  placeholder?: string;
  value: string;
  required?: boolean;
  errorText?: string;
  visible?: boolean;
  disabled?: boolean;
  tooltip?: string;
  inline?: boolean;
  info?: string;
  className?: string;
  validate?: (value: any) => boolean;
  onChange?: (value: string) => void;
};

type NumberProps = {
  label: string;
  type: 'number';
  size?: 'xs' | 'sm' | 'md' | 'xl';
  placeholder?: string;
  value: number;
  required?: boolean;
  errorText?: string;
  visible?: boolean;
  disabled?: boolean;
  tooltip?: string;
  info?: string;
  inline?: boolean;
  validate?: (value: any) => boolean;
  onChange?: (value: number) => void;
  onKeyDown?: (event: React.KeyboardEvent<HTMLInputElement>) => void;
  max?: number;
  min?: number;
};

type CheckboxProps = {
  label?: string;
  type: 'checkbox';
  options: Array<string>;
  value: Array<string>;
  tooltip?: string;
  onChange?: (value: Array<string>) => void;
};

type BooleanProps = {
  label: string;
  type: 'boolean';
  options?: any;
  value: boolean;
  tooltip?: string;
  onChange?: (value: boolean) => void;
};

type RadioProps = {
  label?: string;
  type: 'radio';
  options: Array<string>;
  value: string;
  tooltip?: string;
  onChange?: (value: string) => void;
};

type DropdownProps = {
  label?: string;
  overrideRender?: React.ReactNode;
  type: 'dropdown';
  size?: 'xs' | 'sm' | 'md' | 'xl';
  tooltip?: string;
  required?: boolean;
  inline?: boolean;
  buttonVariant?: 'primary' | 'outline' | 'ghost';
  info?: string;
  disabled?: boolean;
  className?: string;
  value?: number | string;
  options: number[] | string[];
  onChange?: (value: number | string) => void;
};

export type InputProps = TextProps | SliderProps | CheckboxProps | DropdownProps | RadioProps | BooleanProps | NumberProps;

export const Input = (props: InputProps) => {
  switch (props.type) {
    case 'slider':
      return <SliderInput {...(props as SliderProps)} />;
    case 'text' || 'password':
      return <TextInput {...(props as TextProps)} />;
    case 'checkbox':
      return <CheckboxInput {...(props as CheckboxProps)} />;
    case 'dropdown':
      return <DropdownInput {...(props as DropdownProps)} />;
    case 'radio':
      return <RadioInput {...(props as RadioProps)} />;
    case 'boolean':
      return <BooleanInput {...(props as BooleanProps)} />;
    case 'number':
      return <NumberInput {...(props as NumberProps)} />;
    default:
      return null;
  }
};

export const SliderInput = ({ label, value, min, max, step, unit, showValue = true, onChange, tooltip }: SliderProps) => {
  return (
    <Tooltip>
      <TooltipTrigger>
        <div className={styles['slider']}>
          <label className="label flex flex-row justify-between items-end">
            <span className="label-text">{label}</span>
            {showValue ? (
              <div className="label-text">
                {value}
                {unit}
              </div>
            ) : null}
          </label>

          <input
            type="range"
            min={min}
            max={max}
            step={step}
            value={value}
            onChange={(e) => {
              if (onChange) {
                onChange(parseFloat(e.target.value));
              }
            }}
            aria-labelledby="input-slider"
          />
        </div>
      </TooltipTrigger>
      {tooltip && <TooltipContent side={'top'}>{tooltip}</TooltipContent>}
    </Tooltip>
  );
};

export const TextInput = ({
  label,
  placeholder,
  value = '',
  size = 'md',
  required = false,
  visible = true,
  errorText,
  validate,
  disabled = false,
  onChange,
  inline = false,
  tooltip,
  info,
  className,
}: TextProps) => {
  const [isValid, setIsValid] = React.useState<boolean>(true);

  if (!tooltip && inline) tooltip = label;
  return (
    <Tooltip>
      <TooltipTrigger className={styles['input'] + ' form-control w-full' + (inline ? ' grid grid-cols-2 items-center gap-0.5' : '')}>
        {label && (
          <label className="label p-0">
            <span
              className={`text-sm font-medium text-secondary-700 ${required && "after:content-['*'] after:ml-0.5 after:text-danger-500"}`}
            >
              {label}
            </span>
            {required && isValid ? null : <span className="label-text-alt text-error">{errorText}</span>}
            {info && <Info tooltip={info} side={'left'} />}
          </label>
        )}
        <input
          type={visible ? 'text' : 'password'}
          placeholder={placeholder}
          className={
            `${size} bg-light border border-secondary-200 placeholder-secondary-400 focus:outline-none block w-full focus:ring-1 ${
              isValid ? '' : 'input-error'
            }` + (className ? ` ${className}` : '')
          }
          value={value.toString()}
          onChange={(e) => {
            if (required && validate) {
              setIsValid(validate(e.target.value));
            }
            if (onChange) {
              onChange(e.target.value);
            }
          }}
          required={required}
          disabled={disabled}
        />
      </TooltipTrigger>
      {tooltip && <TooltipContent side={'top'}>{tooltip}</TooltipContent>}
    </Tooltip>
  );
};

export const NumberInput = ({
  label,
  placeholder,
  size = 'md',
  value = 0,
  inline = false,
  required = false,
  visible = true,
  errorText,
  validate,
  disabled = false,
  onChange,
  tooltip,
  info,
  onKeyDown,
  max,
  min,
}: NumberProps) => {
  const [isValid, setIsValid] = React.useState<boolean>(true);

  if (!tooltip && inline) tooltip = label;
  return (
    <Tooltip>
      <TooltipTrigger className={styles['input'] + ' form-control w-full' + (inline ? ' grid grid-cols-2 items-center gap-0.5' : '')}>
        <label className="label p-0">
          <span
            className={`text-sm text-left truncate font-medium text-secondary-700 ${required && "after:content-['*'] after:ml-0.5 after:text-danger-500"}`}
          >
            {label}
          </span>
          {required && isValid ? null : <span className="label-text-alt text-error">{errorText}</span>}
          {info && <Info tooltip={info} side={'left'} />}
        </label>
        <input
          type="number"
          placeholder={placeholder}
          className={`${size} bg-light border border-secondary-200 placeholder-secondary-400 focus:outline-none block w-full sm:text-sm focus:ring-1 ${
            isValid ? '' : 'input-error'
          }`}
          value={value.toString()}
          onChange={(e) => {
            if (required && validate) {
              setIsValid(validate(e.target.value));
            }
            if (onChange) {
              onChange(Number(e.target.value));
            }
          }}
          required={required}
          disabled={disabled}
          onKeyDown={onKeyDown}
          max={max}
          min={min}
        />
      </TooltipTrigger>
      {tooltip && <TooltipContent side={'top'}>{tooltip}</TooltipContent>}
    </Tooltip>
  );
};

export const RadioInput = ({ label, value, options, onChange, tooltip }: RadioProps) => {
  return (
    <Tooltip>
      <TooltipTrigger>
        <label className="label p-0">
          <span className="label-text">{label}</span>
        </label>
        {options.map((option, index) => (
          <label key={index} className="label cursor-pointer w-fit gap-2 px-0 py-1">
            <input
              type="radio"
              name={option}
              checked={value === option}
              onChange={() => {
                if (onChange) {
                  onChange(option);
                }
              }}
              className="radio radio-xs radio-primary"
            />
            <span className="label-text">{option}</span>
          </label>
        ))}
      </TooltipTrigger>
      {tooltip && <TooltipContent side={'top'}>{tooltip}</TooltipContent>}
    </Tooltip>
  );
};

export const CheckboxInput = ({ label, value, options, onChange, tooltip }: CheckboxProps) => {
  return (
    <Tooltip>
      <TooltipTrigger>
        {label && (
          <label className="label p-0">
            <span className="label-text">{label}</span>
          </label>
        )}
        {options.map((option, index) => (
          <label key={index} className="label cursor-pointer w-fit gap-2 px-0 py-1">
            <input
              type="checkbox"
              name={option}
              checked={Array.isArray(value) && value.includes(option)}
              onClick={(event) => {
                if (event.ctrlKey) {
                  const updatedValue = (event.target as any).checked ? [option] : value.filter((val) => val !== option);
                  if (onChange) {
                    onChange(updatedValue);
                  }
                } else {
                  const updatedValue = (event.target as any).checked ? [...value, option] : value.filter((val) => val !== option);
                  if (onChange) {
                    onChange(updatedValue);
                  }
                }
              }}
              onChange={() => {
                // to remove warning
              }}
              className="checkbox checkbox-xs"
            />
            <span className="label-text">{option}</span>
          </label>
        ))}
      </TooltipTrigger>
      {tooltip && <TooltipContent side={'top'}>{tooltip}</TooltipContent>}
    </Tooltip>
  );
};

export const BooleanInput = ({ label, value, onChange, tooltip }: BooleanProps) => {
  return (
    <Tooltip>
      <TooltipTrigger>
        <label className={`label cursor-pointer w-fit gap-2 px-0 py-1`}>
          <span className="text-sm">{label}</span>
          <input
            type="checkbox"
            checked={value}
            onChange={(event) => {
              if (onChange) {
                onChange(event.target.checked);
              }
            }}
            className="checkbox checkbox-xs"
          />
        </label>
      </TooltipTrigger>
      {tooltip && <TooltipContent side={'top'}>{tooltip}</TooltipContent>}
    </Tooltip>
  );
};

export const DropdownInput = ({
  label,
  value,
  overrideRender,
  options,
  onChange,
  required = false,
  tooltip,
  size = 'sm',
  inline = true,
  disabled = false,
  buttonVariant = 'primary',
  className = '',
  info,
}: DropdownProps) => {
  const [isDropdownOpen, setIsDropdownOpen] = React.useState<boolean>(false);

  if (!tooltip && inline) tooltip = label;

  return (
    <Tooltip>
      <TooltipTrigger className={className + ' w-full' + (inline ? ' grid grid-cols-2 items-center gap-0.5' : '')}>
        {label && (
          <label className="label p-0">
            <span
              className={`text-${size} text-left truncate font-medium text-secondary-700 ${required && "after:content-['*'] after:ml-0.5 after:text-danger-500"}`}
            >
              {label}
            </span>
            {info && <Info tooltip={info} />}
          </label>
        )}
        <DropdownContainer
          open={isDropdownOpen}
          onOpenChange={(ref) => {
            setIsDropdownOpen(ref);
          }}
        >
          <DropdownTrigger
            variant={buttonVariant}
            title={overrideRender || value}
            size={size}
            className="cursor-pointer"
            disabled={disabled}
            onClick={() => {
              setIsDropdownOpen(!isDropdownOpen);
            }}
          />
          {isDropdownOpen && (
            <DropdownItemContainer className="w-fit">
              {options &&
                options.map((item: any, index: number) => (
                  <DropdownItem
                    key={index}
                    value={item.toString()}
                    selected={value === item}
                    onClick={() => {
                      const parsedValue = typeof item === 'number' ? item.toString() : item;
                      if (onChange) {
                        onChange(parsedValue);
                      }
                      setIsDropdownOpen(false);
                    }}
                  />
                ))}
            </DropdownItemContainer>
          )}
        </DropdownContainer>
      </TooltipTrigger>
      {tooltip && <TooltipContent side={'top'}>{tooltip}</TooltipContent>}
    </Tooltip>
  );
};
