import React, { useEffect, useRef, useState } from 'react';
import { select, scaleBand, axisBottom, scaleLinear, forceX, forceY, brush, forceCollide, format, axisLeft, forceSimulation } from 'd3';

import { VisualRegionConfig, RegionData, DataPoint, DataPointXY } from './types';
import { calcTextWidth, calcTextWidthCanvas } from './utils';
import { ArrowRightAlt } from '@mui/icons-material';
import Icon from '@graphpolaris/shared/lib/components/icon';

export type ScatterPlotProps = {
  data: RegionData;
  visualConfig: VisualRegionConfig;
  xScaleRange: string[] | number[];
  yScaleRange: string[] | number[];
  width: number;
  onBrushUpdate: (idElements: string[], selectedElement: string) => void;
  onBrushClear: (selectedElement: string, idData: string[]) => void;
  onResultJitter: (data: DataPoint[]) => void;
};

export type KeyedScatterplotProps = ScatterPlotProps & {
  key: number;
};

function computeRadiusPoints(width: number, numPoints: number): number {
  const radius: number = numPoints >= 170 ? width * 0.0042 : width * 0.007;

  return radius;
}

export const ScatterPlot = ({
  data,
  visualConfig,
  xScaleRange,
  yScaleRange,
  onBrushUpdate,
  onBrushClear,
  onResultJitter,
  width,
}: ScatterPlotProps) => {
  const svgRef = useRef(null);
  const groupMarginRef = useRef<SVGGElement>(null as any);

  const brushRef = useRef<SVGGElement>(null as any);

  useEffect(() => {
    const maxLengthAllowedAxisY: number = 85;
    const maxLengthAllowedAxisX: number = 50;
    const styleTextXAxisLabel = {
      classTextXAxis: 'font-inter font-secondary font-semibold text-right capitalize text-xs',
      x: 1.01 * visualConfig.widthMargin + visualConfig.margin.right,
      y: visualConfig.heightMargin + 1.25 * visualConfig.margin.bottom,
      textAnchor: 'start',
      dominantBaseline: 'middle',
      maxLengthText: 90,
    };

    const svg = select(svgRef.current);

    const groupMargin = select(groupMarginRef.current).attr(
      'transform',
      `translate(${visualConfig.margin.left},${visualConfig.margin.top})`,
    );

    const maxComputations = 200;
    let tickCount = 0;
    let dataCircles: DataPointXY[] = [];

    let dataCirclesXTemp: number[] = [];
    let dataCirclesYTemp: number[] = [];
    let xOffset: number;

    if (data.xData.length != 0 && data.yData.length != 0) {
      if (typeof data.xData[0] != 'number') {
        let xScaleTemp = scaleBand<string>()
          .domain(xScaleRange as string[])
          .range([0, visualConfig.widthMargin])
          .paddingOuter(0);

        xOffset = 0.5 * xScaleTemp.bandwidth();

        dataCirclesXTemp = data.xData.map((value, index) => {
          const scaledValue = typeof value === 'number' ? xScaleTemp(value.toString()) : xScaleTemp(value);
          if (scaledValue !== undefined) {
            return scaledValue + xOffset;
          } else {
            return 0;
          }
        });

        const textTicks = calcTextWidth(xScaleRange as string[], maxLengthAllowedAxisX, styleTextXAxisLabel.classTextXAxis);

        let xAxis = axisBottom(xScaleTemp)
          .tickFormat((d, i) => textTicks[i])
          .tickSizeOuter(0);
        groupMargin
          .append('g')
          .attr('transform', 'translate(0,' + visualConfig.heightMargin + ')')
          .call(xAxis)
          .selectAll('text')
          .style('text-anchor', 'start')
          .attr('x', '10')
          .attr('y', '0')
          .attr('dy', '0')
          .style('dominant-baseline', styleTextXAxisLabel.dominantBaseline)
          .attr('transform', 'rotate(90)');
      } else {
        let xScaleTemp = scaleLinear<number>()
          .domain(xScaleRange as number[])
          .range([0, visualConfig.widthMargin]);

        dataCirclesXTemp = data.xData.map((value, index) => {
          const numericValue = typeof value === 'string' ? parseFloat(value) : value;
          return xScaleTemp(numericValue);
        });

        const [minValueX, maxValueX]: number[] = xScaleTemp.domain();
        const averageMinMaxX: number = Math.round((minValueX + maxValueX) / 2.0);

        const xAxis = axisBottom(xScaleTemp)
          .tickValues([minValueX, 0.5 * (minValueX + averageMinMaxX), averageMinMaxX, 0.5 * (maxValueX + averageMinMaxX), maxValueX])
          .tickFormat(format('.2s'));

        groupMargin
          .append('g')
          .attr('transform', 'translate(0,' + visualConfig.heightMargin + ')')
          .call(xAxis);
      }

      if (typeof data.yData[0] != 'number') {
        let yScaleTemp = scaleBand<string>()
          .domain(yScaleRange as string[])
          .range([visualConfig.heightMargin, 0])
          .paddingOuter(0);

        xOffset = 0.5 * yScaleTemp.bandwidth();

        dataCirclesYTemp = data.yData.map((value, index) => {
          const scaledValue = typeof value === 'number' ? yScaleTemp(value.toString()) : yScaleTemp(value);
          if (scaledValue !== undefined) {
            return scaledValue + xOffset;
          } else {
            return 0;
          }
        });

        const textTicks = calcTextWidth(yScaleRange as string[], maxLengthAllowedAxisY, styleTextXAxisLabel.classTextXAxis);

        let yAxis = axisLeft(yScaleTemp)
          .tickFormat((d, i) => textTicks[i])
          .tickSizeOuter(0);
        groupMargin.append('g').call(yAxis).selectAll('text');
      } else {
        let yScaleTemp = scaleLinear<number>()
          .domain(yScaleRange as number[])
          .range([visualConfig.heightMargin, 0]);

        dataCirclesYTemp = data.yData.map((value, index) => {
          const numericValue = typeof value === 'string' ? parseFloat(value) : value;
          return yScaleTemp(numericValue);
        });

        const [minValueX, maxValueX]: number[] = yScaleTemp.domain();
        const averageMinMaxX: number = Math.round((minValueX + maxValueX) / 2.0);

        const yAxis = axisLeft(yScaleTemp)
          .tickValues([minValueX, 0.5 * (minValueX + averageMinMaxX), averageMinMaxX, 0.5 * (maxValueX + averageMinMaxX), maxValueX])
          .tickFormat(format('.2s'));

        groupMargin.append('g').call(yAxis);
      }

      dataCircles = data.xData.map((value, index) => ({ x: dataCirclesXTemp[index], y: dataCirclesYTemp[index] }));

      const radius = computeRadiusPoints(visualConfig.width, data.xData.length);
      groupMargin
        .selectAll('circle')
        .data(dataCircles)
        .enter()
        .append('circle')
        .attr('class', (d, i) => `${data.idData[i]}`)
        .attr('cx', (d) => d.x)
        .attr('cy', (d) => d.y)
        .attr('r', radius)
        .attr('stroke', data.colorNodesStroke)
        .attr('fill', data.colorNodes);

      const dataSimulation: DataPoint[] = dataCircles.map(({ x, y }, i) => ({
        x: x,
        y: y,
        id: data.idData[i],
      }));

      onResultJitter(dataSimulation);

      const textLabelAxis = calcTextWidth(data.xAxisName, styleTextXAxisLabel.maxLengthText, styleTextXAxisLabel.classTextXAxis);

      svg
        .append('text')
        .attr('x', styleTextXAxisLabel.x)
        .attr('y', styleTextXAxisLabel.y)
        .text(textLabelAxis[0])
        .style('text-anchor', styleTextXAxisLabel.textAnchor)
        .style('dominant-baseline', styleTextXAxisLabel.dominantBaseline)
        .attr('class', styleTextXAxisLabel.classTextXAxis);
    } else if (data.yData.length != 0) {
      if (typeof data.yData[0] != 'number') {
        let yScaleTemp = scaleBand<string>()
          .domain(yScaleRange as string[])
          .range([visualConfig.heightMargin, 0])
          .paddingOuter(0);

        xOffset = 0.5 * yScaleTemp.bandwidth();

        dataCirclesYTemp = data.yData.map((value, index) => {
          const scaledValue = typeof value === 'number' ? yScaleTemp(value.toString()) : yScaleTemp(value);
          if (scaledValue !== undefined) {
            return scaledValue + xOffset;
          } else {
            return 0;
          }
        });

        const textTicks = calcTextWidth(yScaleRange as string[], maxLengthAllowedAxisY, styleTextXAxisLabel.classTextXAxis);

        let yAxis = axisLeft(yScaleTemp)
          .tickFormat((d, i) => textTicks[i])
          .tickSizeOuter(0);
        groupMargin.append('g').call(yAxis).selectAll('text');
      } else {
        let yScaleTemp = scaleLinear<number>()
          .domain(yScaleRange as number[])
          .range([visualConfig.heightMargin, 0]);

        dataCirclesYTemp = data.yData.map((value, index) => {
          const numericValue = typeof value === 'string' ? parseFloat(value) : value;
          return yScaleTemp(numericValue);
        });

        const [minValueX, maxValueX]: number[] = yScaleTemp.domain();
        const averageMinMaxX: number = Math.round((minValueX + maxValueX) / 2.0);

        const xAxis = axisLeft(yScaleTemp)
          .tickValues([minValueX, 0.5 * (minValueX + averageMinMaxX), averageMinMaxX, 0.5 * (maxValueX + averageMinMaxX), maxValueX])
          .tickFormat(format('.2s'));

        groupMargin.append('g').call(xAxis);
      }

      let xScaleTemp = scaleLinear<number>()
        .domain(xScaleRange as number[])
        .range([0, visualConfig.widthMargin]);

      dataCircles = data.yData.map((value, index) => ({ x: xScaleTemp(0.0), y: dataCirclesYTemp[index] }));

      const radius = computeRadiusPoints(visualConfig.width, data.yData.length);
      const simulation = forceSimulation<DataPointXY>(dataCircles)
        .force('x', forceX<DataPointXY>((d) => d.x).strength(0.1))
        .force('y', forceY<DataPointXY>((d) => d.y).strength(4))
        .force('collide', forceCollide(radius * 1.25).strength(0.5));

      const circles = groupMargin
        .selectAll('circle')
        .data(dataCircles)
        .enter()
        .append('circle')
        .attr('class', (d, i) => `${data.idData[i]}`)
        .attr('cx', (d) => d.x)
        .attr('cy', (d) => d.y)
        .attr('r', radius)
        .attr('stroke', data.colorNodesStroke)
        .attr('fill', data.colorNodes);

      simulation.on('tick', function () {
        circles.attr('cx', (d, i) => d.x as number).attr('cy', (d) => d.y as number);

        tickCount++;
        if (tickCount > maxComputations) {
          const dataSimulation: DataPoint[] = dataCircles.map(({ x, y }, i) => ({
            x,
            y,
            id: data.idData[i],
          }));
          onResultJitter(dataSimulation);

          simulation.stop();
        }
      });
    } else if (data.xData.length != 0) {
      if (typeof data.xData[0] != 'number') {
        let xScaleTemp = scaleBand<string>()
          .domain(xScaleRange as string[])
          .range([0, visualConfig.widthMargin])
          .paddingOuter(0);

        xOffset = 0.5 * xScaleTemp.bandwidth();

        dataCirclesXTemp = data.xData.map((value, index) => {
          const scaledValue = typeof value === 'number' ? xScaleTemp(value.toString()) : xScaleTemp(value);
          if (scaledValue !== undefined) {
            return scaledValue + xOffset;
          } else {
            return 0;
          }
        });

        const textTicks = calcTextWidth(xScaleRange as string[], maxLengthAllowedAxisX, styleTextXAxisLabel.classTextXAxis);

        let xAxis = axisBottom(xScaleTemp)
          .tickFormat((d, i) => textTicks[i])
          .tickSizeOuter(0);
        groupMargin
          .append('g')
          .attr('transform', 'translate(0,' + visualConfig.heightMargin + ')')
          .call(xAxis)
          .selectAll('text')
          .style('text-anchor', 'start')
          .attr('x', '10')
          .attr('y', '0')
          .attr('dy', '0')
          .style('dominant-baseline', styleTextXAxisLabel.dominantBaseline)
          .attr('transform', 'rotate(90)');
      } else {
        let xScaleTemp = scaleLinear<number>()
          .domain(xScaleRange as number[])
          .range([0, visualConfig.widthMargin]);

        dataCirclesXTemp = data.xData.map((value, index) => {
          const numericValue = typeof value === 'string' ? parseFloat(value) : value;
          return xScaleTemp(numericValue);
        });

        const [minValueX, maxValueX]: number[] = xScaleTemp.domain();
        const averageMinMaxX: number = Math.round((minValueX + maxValueX) / 2.0);

        const xAxis = axisBottom(xScaleTemp)
          .tickValues([minValueX, 0.5 * (minValueX + averageMinMaxX), averageMinMaxX, 0.5 * (maxValueX + averageMinMaxX), maxValueX])
          .tickFormat(format('.2s'));

        groupMargin
          .append('g')
          .attr('transform', 'translate(0,' + visualConfig.heightMargin + ')')
          .call(xAxis);
      }

      let yScaleTemp = scaleLinear<number>()
        .domain(yScaleRange as number[])
        .range([visualConfig.heightMargin, 0]);

      dataCircles = data.xData.map((value, index) => ({ x: dataCirclesXTemp[index], y: yScaleTemp(0) }));

      const radius = computeRadiusPoints(visualConfig.width, data.xData.length);
      const simulation = forceSimulation<DataPointXY>(dataCircles)
        .force('x', forceX<DataPointXY>((d) => d.x).strength(4))
        .force('y', forceY<DataPointXY>((d) => d.y).strength(0.1))
        .force('collide', forceCollide(radius * 1.25).strength(0.5));

      const circles = groupMargin
        .selectAll('circle')
        .data(dataCircles)
        .enter()
        .append('circle')
        .attr('class', (d, i) => `${data.idData[i]}`)
        .attr('cx', (d) => d.x || 0)
        .attr('cy', (d) => d.y || 0)
        .attr('r', radius)
        .attr('stroke', data.colorNodesStroke)
        .attr('fill', data.colorNodes);

      simulation.on('tick', function () {
        circles.attr('cx', (d, i) => d.x as number).attr('cy', (d) => d.y as number);

        tickCount++;
        if (tickCount > maxComputations) {
          const dataSimulation: DataPoint[] = dataCircles.map(({ x, y }, i) => ({
            x,
            y,
            id: data.idData[i],
          }));

          onResultJitter(dataSimulation);

          simulation.stop();
        }
      });
      const textLabelAxis = calcTextWidth(data.xAxisName, styleTextXAxisLabel.maxLengthText, styleTextXAxisLabel.classTextXAxis);

      svg
        .append('text')
        .attr('x', styleTextXAxisLabel.x)
        .attr('y', styleTextXAxisLabel.y)
        .text(textLabelAxis[0])
        .style('text-anchor', styleTextXAxisLabel.textAnchor)
        .style('dominant-baseline', styleTextXAxisLabel.dominantBaseline)
        .attr('class', styleTextXAxisLabel.classTextXAxis);
    }

    svg
      .append('rect')
      .attr('x', 0)
      .attr('y', 0)
      .attr('width', visualConfig.width)
      .attr('height', visualConfig.height)
      .attr('fill', 'none')
      .attr('stroke', 'gray');

    svg.selectAll('.domain').style('stroke', 'hsl(var(--clr-sec--400))');
    svg.selectAll('.tick line').style('stroke', 'hsl(var(--clr-sec--400))');
    svg.selectAll('.tick text').attr('class', 'font-mono').style('stroke', 'none').style('fill', 'hsl(var(--clr-sec--500))');

    // BRUSH LOGIC
    const myBrush: any = brush()
      .extent([
        [0, 0],
        [visualConfig.widthMargin, visualConfig.heightMargin],
      ])
      .on('brush', brushed)
      .on('end', function (event: any) {
        if (event.selection === null) {
          onBrushClear(data.name, data.idData);
        }
      });

    let selectedDataIds: string[] = [];

    function brushed(event: any) {
      if (!event.selection) {
      } else {
        const [[x0, y0], [x1, y1]] = event.selection;
        dataCircles.forEach((d, i) => {
          if (d.x >= x0 && d.x <= x1 && d.y >= y0 && d.y <= y1) {
            selectedDataIds.push(data.idData[i]);
          }
        });

        onBrushUpdate(selectedDataIds, data.name);
        selectedDataIds = [];
      }
    }

    const brushGroup = select(brushRef.current);
    brushGroup.attr('transform', `translate(${visualConfig.margin.left},${visualConfig.margin.top})`).attr('class', 'brushingElem');

    brushGroup.call(myBrush);

    brushGroup
      .select('.selection')
      .style('fill', data.colorBrush)
      .style('fill-opacity', 0.5)
      .style('stroke', data.colorBrushStroke)
      .style('stroke-width', 2);

    brushGroup.selectAll('.handle').style('stroke', 'none');
  }, [data, visualConfig, xScaleRange, yScaleRange, width]);

  return (
    <div className="w-full border border-secondary-200">
      <div className="w-full border-light font-secondary group bg-secondary-200 truncate text-left absolute">
        <div className="mx-1 w-full flex items-center">
          <span>{data.nodeName}</span> <Icon component={<ArrowRightAlt />} size={32} />
          <span>{data.attributeName}</span> <span> : </span>
          <span>{data.attributeSelected}</span>
        </div>
      </div>
      <div className="w-full flex flex-row justify-center">
        <svg ref={svgRef} width={600} height={visualConfig.height}>
          <g ref={groupMarginRef} />
          <g ref={brushRef} />
        </svg>
      </div>
    </div>
  );
};
